import re

from django.db.models import Q

valid_time_offset_regex = re.compile(r"^\d+[usmhdw]$")


def is_time_offset_valid(offset):
    """ensures that the given offset is valid to influxdb

    :param offset: the offset string
    :type offset: str

    :return: is this a valid offset
    :rtype: bool
    """
    return re.match(valid_time_offset_regex, offset) is not None


def get_django_instance_from_point(method, field, point):
    """gets a django model instance via the ORM from a point's content id

    :param method: a django model manager method to get an instance
    :type method: method

    :param field: the name of the django model field that corresponds to a content id
    :type field: str

    :param point: a point object from influxdb
    :type point: Point

    :return: a django model instance
    :rtype: Model
    """
    query = {field: point.content_id}
    return method(Q(**query))



from .clients import InfluxerClient


class DummyClient(InfluxerClient):
    """a dummy client that returns an empty list to `get_content` - useful only for travis testing
    """

    @classmethod
    def from_real_client(cls, real_client):
        """

        :param real_client:
        :return:
        """
        return cls(
            real_client.host, real_client.port, real_client.username, real_client.password, real_client.database,
            real_client.series, real_client.ssl, real_client.timeout
        )

    def get_content(self, offset, limit=None):
        return []



from influxdb import InfluxDBClient

from .models import Point
from .utils import is_time_offset_valid


class InfluxerClient(object):

    def __init__(self, host, port, username, password, database, series, ssl=False, timeout=None):
        """initializes a client

        :param host: the host to connect to - can be a hostname or an ip address
        :type host: str

        :param port: the port number to connect on the host
        :type port: int

        :param username: the username to connect with - should have at least read access
        :type username: str

        :param password: the password for the username
        :param password: str

        :param database: the name of the database to connect to
        :type database: str

        :param series: the name of the series to read from
        :type series: str

        :param ssl: should we connect using ssl
        :type ssl: bool

        :param timeout: time it takes to drop a connection
        :type timeout: int
        """
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.database = database
        self.series = series
        self.ssl = ssl
        self.timeout = timeout
        self._client = InfluxDBClient(
            self.host, self.port, self.username, self.password, self.database, ssl=self.ssl, timeout=self.timeout)

    def get_content(self, offset, limit=None):
        """gets content from the client's series

        :param offset: the time offset to query by
        :type offset: str

        :param limit: cap on the number of results
        :type limit: int

        :return: data points that match the offset capped by the limit
        :rtype: list
        """
        if not is_time_offset_valid(offset):
            raise Exception("The offset provided is not valid.")

        query = "SELECT SUM(value) AS value " \
                "FROM {series} " \
                "WHERE time > now() - {offset} " \
                "GROUP BY content_id;".format(series=self.series, offset=offset)
        results = self._client.query(query).raw

        points = []
        results = results.get("series", [])
        for result in results:
            name = result.get("name")
            cols = result.get("columns", [])
            vals = result.get("values", [])
            cid = result.get("tags", {}).get("content_id")
            cols.append("content_id")
            for val in vals:
                val.append(cid)
                obj = dict(zip(cols, val))
                obj["site"] = name
                points.append(Point(**obj))

        return sorted(points, key=lambda p: p.value, reverse=True)



"""
objects to handle stored points in influxdb
"""


class Point(object):
    """
    a piece of information stored in influxdb
    """

    def __init__(self, site, content_id, value, event=None, path=None, time=None, sequence_number=None):
        self.site = site
        self.content_id = content_id
        self.value = value
        self.event = event
        self.path = path
        self.time = time
        self.sequence_number = sequence_number

    def __str__(self):
        return "<Point: content_id={} value={}>".format(self.content_id, self.value)

    def __repr__(self):
        return str(self)

    def __lt__(self, other):
        """compares this instance to another based on their `value` attributes

        :param other: the other Point object to compare
        :type other: Point

        :return: is self.value less than the other value
        :rtype: bool
        """
        return self.value < other.value



from distutils.core import setup


setup(
    name="django-influxer-client",
    version="2.0.2",
    description="Modules and JS to work with Django-Influxer integration",
    author="Vince Forgione",
    author_email="vforgione@theonion.com",
    url="https://github.com/theonion/django-influxer-client",
    packages=["djinfluxer"],
    install_requires=[
        "influxdb",
    ]
)



from sys import maxint

from django.contrib.auth import get_user_model
from django.test import TestCase
from requests.exceptions import ConnectionError

from .clients import InfluxerClient
from .dummy import DummyClient
from .models import Point
from .utils import is_time_offset_valid, get_django_instance_from_point


class PointTests(TestCase):
    """tests methods of the Point object
    """

    def setUp(self):
        """create two, non-equal point objects
        """
        self.point1 = Point("onion", 123, 10, "pageview", "/articles/read-this-123")
        self.point2 = Point("onion", 456, 21, "pageview", "/articles/read-this-too-456")

    def test__lt(self):
        """tests the __lt__ method of Point
        """
        self.assertLess(self.point1, self.point2)


class ClientTests(TestCase):
    """tests methods of the InfluxerClient object
    """

    def setUp(self):
        """sets up a client to an influxdb service at root:root@localhost:8086
        """
        self.client = InfluxerClient("127.0.0.1", 8086, "root", "root", "influxdb", "test")

    def test__get_content(self):
        """tests the get_content method of InfluxerClient - will skip if connection errs
        """
        value = maxint
        content_ids = []
        try:
            points = self.client.get_content("15m", limit=25)
            for point in points:
                self.assertIsInstance(point, Point)
                self.assertLessEqual(point.value, value)
                self.assertNotIn(point.content_id, content_ids)
                value = point.value
                content_ids.append(point.content_id)
        except ConnectionError:
            return


class DummyClientTests(TestCase):
    """tests methods of the DummyClient object
    """

    def setUp(self):
        """sets up a client to an influxdb service at root:root@localhost:8086
        """
        self.client = DummyClient("127.0.0.1", 8086, "root", "root", "influxdb", "test")

    def test__from_real_client(self):
        client = InfluxerClient("127.0.0.1", 8086, "root", "root", "influxdb", "test")
        dummy = DummyClient.from_real_client(client)
        self.assertIsInstance(dummy, DummyClient)
        self.assertTrue(hasattr(dummy, "get_content"))

    def test__get_content(self):
        """tests the get_content method returns an empty list
        """
        points = self.client.get_content("1d")
        self.assertEqual(points, [])
        # test dummy client from `from_real_client` constructor
        client = InfluxerClient("127.0.0.1", 8086, "root", "root", "influxdb", "test")
        dummy = DummyClient.from_real_client(client)
        points = dummy.get_content("1d")
        self.assertEqual(points, [])


class UtilTests(TestCase):
    """tests functions in the utils module
    """

    def test__is_time_offset_valid(self):
        """tests is_time_offset_valid
        """
        valid_offsets = ("u", "s", "m", "h", "d", "w")
        invalid_offsets = ("y", "microsecond", "minute")
        for index, offset in enumerate(valid_offsets):
            good = "{}{}".format(index, offset)
            self.assertTrue(is_time_offset_valid(good))
            bad = "a{}".format(offset)
            self.assertFalse(is_time_offset_valid(bad))
        for index, offset in enumerate(invalid_offsets):
            bad = "{}{}".format(index, offset)
            self.assertFalse(is_time_offset_valid(bad))

    def test__get_django_instance_from_point(self):
        """tests get_django_instance_from_point
        """
        User = get_user_model()
        user = User.objects.create(username="test", password="user", email="test.user@example.com")
        point = Point("onion", user.pk, "whatever, this doesn't matter in this example")
        method = User.objects.get
        field = "pk"
        instance = get_django_instance_from_point(method, field, point)
        self.assertIsInstance(instance, User)
